searchData={"items":[{"type":"module","title":"Poolex","doc":"","ref":"Poolex.html"},{"type":"module","title":"Usage - Poolex","doc":"In the most typical use of Poolex, you only need to start pool of workers as a child of your application.\n\n```elixir\nchildren = [\n  {Poolex,\n    pool_id: :worker_pool,\n    worker_module: SomeWorker,\n    workers_count: 5}\n]\n\nSupervisor.start_link(children, strategy: :one_for_one)\n```\n\nThen you can execute any code on the workers with `run/3`:\n\n```elixir\nPoolex.run(:worker_pool, &(is_pid?(&1)), checkout_timeout: 1_000)\n{:ok, true}\n```\n\nFore more information see [Getting Started](https://hexdocs.pm/poolex/getting-started.html)","ref":"Poolex.html#module-usage"},{"type":"function","title":"Poolex.add_idle_workers!/2","doc":"Adds some idle workers to existing pool.","ref":"Poolex.html#add_idle_workers!/2"},{"type":"function","title":"Poolex.child_spec/1","doc":"Returns a specification to start this module under a supervisor.","ref":"Poolex.html#child_spec/1"},{"type":"function","title":"Options - Poolex.child_spec/1","doc":"| Option                 | Description                                          | Example               | Default value                     |\n|------------------------|------------------------------------------------------|-----------------------|-----------------------------------|\n| `pool_id`              | Identifier by which you will access the pool         | `:my_pool`            | `worker_module` value             |\n| `worker_module`        | Name of module that implements our worker            | `MyApp.Worker`        | **option is required**            |\n| `workers_count`        | How many workers should be running in the pool       | `5`                   | **option is required**            |\n| `max_overflow`         | How many workers can be created over the limit       | `2`                   | `0`                               |\n| `worker_args`          | List of arguments passed to the start function       | `[:gg, \"wp\"]`         | `[]`                              |\n| `worker_start_fun`     | Name of the function that starts the worker          | `:run`                | `:start_link`                     |\n| `busy_workers_impl`    | Module that describes how to work with busy workers  | `SomeBusyWorkersImpl` | `Poolex.Workers.Impl.List`        |\n| `idle_workers_impl`    | Module that describes how to work with idle workers  | `SomeIdleWorkersImpl` | `Poolex.Workers.Impl.List`        |\n| `waiting_callers_impl` | Module that describes how to work with callers queue | `WaitingCallersImpl`  | `Poolex.Callers.Impl.ErlangQueue` |\n| `pool_size_metrics`    | Whether to dispatch pool size metrics                | `true`                | `false`                           |","ref":"Poolex.html#child_spec/1-options"},{"type":"function","title":"Examples - Poolex.child_spec/1","doc":"children = [\n      Poolex.child_spec(worker_module: SomeWorker, workers_count: 5),\n      # or in another way\n      {Poolex, worker_module: SomeOtherWorker, workers_count: 5}\n    ]\n\n    Supervisor.start_link(children, strategy: :one_for_one)","ref":"Poolex.html#child_spec/1-examples"},{"type":"function","title":"Poolex.get_debug_info/1","doc":"Returns detailed information about started pool.\n\nPrimarily needed to help with debugging. **Avoid using this function in production.**","ref":"Poolex.html#get_debug_info/1"},{"type":"function","title":"Fields - Poolex.get_debug_info/1","doc":"* `busy_workers_count` - how many workers are busy right now.\n    * `busy_workers_pids` - list of busy workers.\n    * `idle_workers_count` - how many workers are ready to work.\n    * `idle_workers_pids` - list of idle workers.\n    * `max_overflow` - how many workers can be created over the limit.\n    * `overflow` - current count of workers launched over limit.\n    * `waiting_caller_pids` - list of callers processes.\n    * `worker_args` - what parameters are used to start the worker.\n    * `worker_module` - name of a module that describes a worker.\n    * `worker_start_fun` - what function is used to start the worker.","ref":"Poolex.html#get_debug_info/1-fields"},{"type":"function","title":"Examples - Poolex.get_debug_info/1","doc":"iex> Poolex.start(pool_id: :my_pool_3, worker_module: Agent, worker_args: [fn -> 0 end], workers_count: 5)\n    iex> debug_info = %Poolex.Private.DebugInfo{} = Poolex.get_debug_info(:my_pool_3)\n    iex> debug_info.busy_workers_count\n    0\n    iex> debug_info.idle_workers_count\n    5","ref":"Poolex.html#get_debug_info/1-examples"},{"type":"function","title":"Poolex.get_pool_id/1","doc":"Returns pool identifier from initialization options.","ref":"Poolex.html#get_pool_id/1"},{"type":"type","title":"Poolex.pool_id/0","doc":"Any valid GenServer's name. It may be an atom like `:some_pool` or a tuple {:via, Registry, {MyApp.Registry, \"pool\"}\nif you want to use Registry.","ref":"Poolex.html#t:pool_id/0"},{"type":"type","title":"Poolex.poolex_option/0","doc":"| Option                 | Description                                          | Example               | Default value                     |\n|------------------------|------------------------------------------------------|-----------------------|-----------------------------------|\n| `pool_id`              | Identifier by which you will access the pool         | `:my_pool`            | `worker_module` value             |\n| `worker_module`        | Name of module that implements our worker            | `MyApp.Worker`        | **option is required**            |\n| `workers_count`        | How many workers should be running in the pool       | `5`                   | **option is required**            |\n| `max_overflow`         | How many workers can be created over the limit       | `2`                   | `0`                               |\n| `worker_args`          | List of arguments passed to the start function       | `[:gg, \"wp\"]`         | `[]`                              |\n| `worker_start_fun`     | Name of the function that starts the worker          | `:run`                | `:start_link`                     |\n| `busy_workers_impl`    | Module that describes how to work with busy workers  | `SomeBusyWorkersImpl` | `Poolex.Workers.Impl.List`        |\n| `idle_workers_impl`    | Module that describes how to work with idle workers  | `SomeIdleWorkersImpl` | `Poolex.Workers.Impl.List`        |\n| `waiting_callers_impl` | Module that describes how to work with callers queue | `WaitingCallersImpl`  | `Poolex.Callers.Impl.ErlangQueue` |\n| `pool_size_metrics`    | Whether to dispatch pool size metrics                | `true`                | `false`                           |","ref":"Poolex.html#t:poolex_option/0"},{"type":"function","title":"Poolex.remove_idle_workers!/2","doc":"Removes some idle workers from existing pool.\nIf the number of workers to remove is greater than the number of idle workers, all idle workers will be removed.","ref":"Poolex.html#remove_idle_workers!/2"},{"type":"function","title":"Poolex.run/3","doc":"The main function for working with the pool.\n\nIt takes a pool identifier, a function that takes a worker process id as an argument and returns any value.\nWhen executed, an attempt is made to find a free worker with specified timeout (5 seconds by default).\nYou can set the timeout using the `checkout_timeout` option.\n\nReturns:\n  * `{:ok, result}` if the worker was found and the function was executed successfully.\n  * `{:error, :checkout_timeout}` if no free worker was found before the timeout.","ref":"Poolex.html#run/3"},{"type":"function","title":"Examples - Poolex.run/3","doc":"iex> Poolex.start_link(pool_id: :some_pool, worker_module: Agent, worker_args: [fn -> 5 end], workers_count: 1)\n    iex> Poolex.run(:some_pool, fn pid -> Agent.get(pid, &(&1)) end)\n    {:ok, 5}","ref":"Poolex.html#run/3-examples"},{"type":"type","title":"Poolex.run_option/0","doc":"| Option  | Description                                        | Example  | Default value                           |\n|---------|----------------------------------------------------|----------|-----------------------------------------|\n| checkout_timeout | How long we can wait for a worker on the call site | `60_000` | `5000` |","ref":"Poolex.html#t:run_option/0"},{"type":"function","title":"Poolex.start/1","doc":"Starts a Poolex process without links (outside of a supervision tree).\n\nSee start_link/1 for more information.","ref":"Poolex.html#start/1"},{"type":"function","title":"Examples - Poolex.start/1","doc":"iex> Poolex.start(pool_id: :my_pool, worker_module: Agent, worker_args: [fn -> 0 end], workers_count: 5)\n    iex> %Poolex.Private.State{worker_module: worker_module} = :sys.get_state(:my_pool)\n    iex> worker_module\n    Agent","ref":"Poolex.html#start/1-examples"},{"type":"function","title":"Poolex.start_link/1","doc":"Starts a Poolex process linked to the current process.\n\nThis is often used to start the Poolex as part of a supervision tree.\n\nAfter the process is started, you can access it using the previously specified `pool_id`.","ref":"Poolex.html#start_link/1"},{"type":"function","title":"Options - Poolex.start_link/1","doc":"| Option                 | Description                                          | Example               | Default value                     |\n|------------------------|------------------------------------------------------|-----------------------|-----------------------------------|\n| `pool_id`              | Identifier by which you will access the pool         | `:my_pool`            | `worker_module` value             |\n| `worker_module`        | Name of module that implements our worker            | `MyApp.Worker`        | **option is required**            |\n| `workers_count`        | How many workers should be running in the pool       | `5`                   | **option is required**            |\n| `max_overflow`         | How many workers can be created over the limit       | `2`                   | `0`                               |\n| `worker_args`          | List of arguments passed to the start function       | `[:gg, \"wp\"]`         | `[]`                              |\n| `worker_start_fun`     | Name of the function that starts the worker          | `:run`                | `:start_link`                     |\n| `busy_workers_impl`    | Module that describes how to work with busy workers  | `SomeBusyWorkersImpl` | `Poolex.Workers.Impl.List`        |\n| `idle_workers_impl`    | Module that describes how to work with idle workers  | `SomeIdleWorkersImpl` | `Poolex.Workers.Impl.List`        |\n| `waiting_callers_impl` | Module that describes how to work with callers queue | `WaitingCallersImpl`  | `Poolex.Callers.Impl.ErlangQueue` |\n| `pool_size_metrics`    | Whether to dispatch pool size metrics                | `true`                | `false`                           |","ref":"Poolex.html#start_link/1-options"},{"type":"function","title":"Examples - Poolex.start_link/1","doc":"iex> Poolex.start_link(pool_id: :other_pool, worker_module: Agent, worker_args: [fn -> 0 end], workers_count: 5)\n    iex> %Poolex.Private.State{worker_module: worker_module} = :sys.get_state(:other_pool)\n    iex> worker_module\n    Agent","ref":"Poolex.html#start_link/1-examples"},{"type":"type","title":"Poolex.worker/0","doc":"Process id of `worker`.\n\n**Workers** are processes launched in a pool.","ref":"Poolex.html#t:worker/0"},{"type":"module","title":"Poolex.Caller","doc":"Caller structure.\n\n**Callers** are processes that have requested to get a worker.","ref":"Poolex.Caller.html"},{"type":"type","title":"Poolex.Caller.t/0","doc":"","ref":"Poolex.Caller.html#t:t/0"},{"type":"behaviour","title":"Poolex.Callers.Behaviour","doc":"Behaviour for callers collection implementations.\n\n`caller` is a process that uses the `Poolex.run/3` function and waits for the execution result.\n\n**Note that the caller's typespec matches `GenServer.from()`**","ref":"Poolex.Callers.Behaviour.html"},{"type":"callback","title":"Poolex.Callers.Behaviour.add/2","doc":"Adds caller to `state` and returns new state.","ref":"Poolex.Callers.Behaviour.html#c:add/2"},{"type":"callback","title":"Poolex.Callers.Behaviour.empty?/1","doc":"Returns `true` if the `state` is empty, `false` otherwise.","ref":"Poolex.Callers.Behaviour.html#c:empty?/1"},{"type":"callback","title":"Poolex.Callers.Behaviour.init/0","doc":"Returns `state` (any data structure) which will be passed as the first argument to all other functions.","ref":"Poolex.Callers.Behaviour.html#c:init/0"},{"type":"callback","title":"Poolex.Callers.Behaviour.pop/1","doc":"Removes one of callers from `state` and returns it as `{caller, state}`. Returns `:empty` if state is empty.","ref":"Poolex.Callers.Behaviour.html#c:pop/1"},{"type":"callback","title":"Poolex.Callers.Behaviour.remove_by_pid/2","doc":"Removes caller by pid from `state` and returns new state.","ref":"Poolex.Callers.Behaviour.html#c:remove_by_pid/2"},{"type":"callback","title":"Poolex.Callers.Behaviour.remove_by_reference/2","doc":"Removes caller by reference from `state` and returns new state.","ref":"Poolex.Callers.Behaviour.html#c:remove_by_reference/2"},{"type":"type","title":"Poolex.Callers.Behaviour.state/0","doc":"","ref":"Poolex.Callers.Behaviour.html#t:state/0"},{"type":"callback","title":"Poolex.Callers.Behaviour.to_list/1","doc":"Returns list of callers.","ref":"Poolex.Callers.Behaviour.html#c:to_list/1"},{"type":"module","title":"Poolex.Callers.Impl.ErlangQueue","doc":"Callers queue (FIFO) implementation based on `:erlang.queue`.","ref":"Poolex.Callers.Impl.ErlangQueue.html"},{"type":"module","title":"Poolex.Private.DebugInfo","doc":"Information with the current state of the pool.\n\nCan be used for debugging.","ref":"Poolex.Private.DebugInfo.html"},{"type":"type","title":"Poolex.Private.DebugInfo.t/0","doc":"","ref":"Poolex.Private.DebugInfo.html#t:t/0"},{"type":"module","title":"Poolex.Private.Metrics","doc":"Functions for dispatching metrics.","ref":"Poolex.Private.Metrics.html"},{"type":"function","title":"Poolex.Private.Metrics.dispatch_pool_size_metrics/1","doc":"Dispatches metrics with current count of idle workers.","ref":"Poolex.Private.Metrics.html#dispatch_pool_size_metrics/1"},{"type":"function","title":"Poolex.Private.Metrics.start_poller/1","doc":"Starts a telemetry poller for dispatching metrics.","ref":"Poolex.Private.Metrics.html#start_poller/1"},{"type":"module","title":"Poolex.Private.State","doc":"Internal structure containing the state of the pool.\n\nCan be used for debugging.","ref":"Poolex.Private.State.html"},{"type":"type","title":"Poolex.Private.State.t/0","doc":"","ref":"Poolex.Private.State.html#t:t/0"},{"type":"behaviour","title":"Poolex.Workers.Behaviour","doc":"Behaviour for worker collection implementations.","ref":"Poolex.Workers.Behaviour.html"},{"type":"callback","title":"Poolex.Workers.Behaviour.add/2","doc":"Adds worker's pid to `state` and returns new state.","ref":"Poolex.Workers.Behaviour.html#c:add/2"},{"type":"callback","title":"Poolex.Workers.Behaviour.count/1","doc":"Returns the number of workers in the state.","ref":"Poolex.Workers.Behaviour.html#c:count/1"},{"type":"callback","title":"Poolex.Workers.Behaviour.empty?/1","doc":"Returns `true` if the `state` is empty, `false` otherwise.","ref":"Poolex.Workers.Behaviour.html#c:empty?/1"},{"type":"callback","title":"Poolex.Workers.Behaviour.init/0","doc":"Returns `state` (any data structure) which will be passed as the first argument to all other functions.","ref":"Poolex.Workers.Behaviour.html#c:init/0"},{"type":"callback","title":"Poolex.Workers.Behaviour.init/1","doc":"Same as `init/0` but returns `state` initialized with passed list of workers.","ref":"Poolex.Workers.Behaviour.html#c:init/1"},{"type":"callback","title":"Poolex.Workers.Behaviour.member?/2","doc":"Returns `true` if given worker contained in the `state`, `false` otherwise.","ref":"Poolex.Workers.Behaviour.html#c:member?/2"},{"type":"callback","title":"Poolex.Workers.Behaviour.pop/1","doc":"Removes one of workers from `state` and returns it as `{caller, state}`. Returns `:empty` if state is empty.","ref":"Poolex.Workers.Behaviour.html#c:pop/1"},{"type":"callback","title":"Poolex.Workers.Behaviour.remove/2","doc":"Removes given worker from `state` and returns new state.","ref":"Poolex.Workers.Behaviour.html#c:remove/2"},{"type":"type","title":"Poolex.Workers.Behaviour.state/0","doc":"","ref":"Poolex.Workers.Behaviour.html#t:state/0"},{"type":"callback","title":"Poolex.Workers.Behaviour.to_list/1","doc":"Returns list of workers pids.","ref":"Poolex.Workers.Behaviour.html#c:to_list/1"},{"type":"module","title":"Poolex.Workers.Impl.ErlangQueue","doc":"Simple workers queue (FIFO) implementation based on Erlang `:queue`","ref":"Poolex.Workers.Impl.ErlangQueue.html"},{"type":"module","title":"Poolex.Workers.Impl.List","doc":"Simple workers stack (LIFO) implementation based on List.","ref":"Poolex.Workers.Impl.List.html"},{"type":"extras","title":"Poolex","doc":"# Poolex\n\n![Build and tests workflow](https://github.com/general-CbIC/poolex/actions/workflows/ci-tests.yml/badge.svg)\n[![hex.pm version](https://img.shields.io/hexpm/v/poolex.svg?style=flat)](https://hex.pm/packages/poolex)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-lightgreen.svg?style=flat)](https://hexdocs.pm/poolex/)\n[![License](https://img.shields.io/hexpm/l/poolex.svg?style=flat)](https://github.com/general-CbIC/poolex/blob/main/LICENSE)\n[![Total Download](https://img.shields.io/hexpm/dt/poolex.svg?style=flat)](https://hex.pm/packages/poolex)\n\nPoolex is a library for managing pools of workers. Inspired by [poolboy](https://github.com/devinus/poolboy).\n\n> [!IMPORTANT]  \n> Documentation on GitHub corresponds to the current branch. For stable versions' docs see [Hexdocs](https://hexdocs.pm/poolex/).","ref":"readme.html"},{"type":"extras","title":"Table of Contents - Poolex","doc":"- [Poolex](#poolex)\n  - [Table of Contents](#table-of-contents)\n  - [Features](#features)\n  - [Requirements](#requirements)\n  - [Installation](#installation)\n  - [Usage](#usage)\n  - [Guides](#guides)\n  - [Contributions](#contributions)","ref":"readme.html#table-of-contents"},{"type":"extras","title":"Features - Poolex","doc":"With `poolex` you can:\n\n- Launch multiple pools of workers and then access the free ones from anywhere in the application.\n- Configure the pool to run additional temporary workers if the load increases.\n- Analyze and optimize your pool's production settings using metrics.\n- Use your implementations to define worker and caller processes access logic.\n\n**Why `poolex` instead of `poolboy`?**\n  \n- `poolex` is written in Elixir. This library is much more convenient to use in Elixir projects.\n- `poolboy` is a great library, but not actively maintained :crying_cat_face:![Last poolboy commit](https://img.shields.io/github/last-commit/devinus/poolboy?style=flat)","ref":"readme.html#features"},{"type":"extras","title":"Requirements - Poolex","doc":"| Requirement | Version |\n|-------------|---------|\n| Erlang/OTP  | >= 22   |\n| Elixir      | >= 1.7  |","ref":"readme.html#requirements"},{"type":"extras","title":"Installation - Poolex","doc":"Add `:poolex` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:poolex, \"~> 1.0\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Usage - Poolex","doc":"In the most typical use of Poolex, you only need to start a pool of workers as a child of your application.\n\n```elixir\nchildren = [\n  {Poolex,\n    worker_module: SomeWorker,\n    workers_count: 5}\n]\n\nSupervisor.start_link(children, strategy: :one_for_one)\n```\n\nThen you can execute any code on the workers with `run/3`:\n\n```elixir\niex> Poolex.run(SomeWorker, &(is_pid?(&1)), checkout_timeout: 1_000)\n{:ok, true}\n```\n\nA detailed description of the available configuration or examples of use can be found in [documentation](https://hexdocs.pm/poolex/getting-started.html).","ref":"readme.html#usage"},{"type":"extras","title":"Guides - Poolex","doc":"- [Getting Started](https://hexdocs.pm/poolex/getting-started.html)\n  - [Starting pool of workers](https://hexdocs.pm/poolex/getting-started.html#starting-pool-of-workers)\n  - [Poolex configuration options](https://hexdocs.pm/poolex/getting-started.html#starting-pool-of-workers)\n  - [Working with the pool](https://hexdocs.pm/poolex/getting-started.html#working-with-the-pool)\n- [Migration from `:poolboy`](https://hexdocs.pm/poolex/migration-from-poolboy.html)\n- [Example of use](https://hexdocs.pm/poolex/example-of-use.html)\n  - [Defining the worker](https://hexdocs.pm/poolex/example-of-use.html#defining-the-worker)\n  - [Configuring Poolex](https://hexdocs.pm/poolex/example-of-use.html#configuring-poolex)\n  - [Using Poolex](https://hexdocs.pm/poolex/example-of-use.html#using-poolex)\n- [Working with metrics](https://hexdocs.pm/poolex/pool-metrics.html)\n  - [Pool size metrics](https://hexdocs.pm/poolex/pool-metrics.html#pool-size-metrics)\n  - [Integration with PromEx](https://hexdocs.pm/poolex/pool-metrics.html#integration-with-promex)\n- [Workers and callers implementations](https://hexdocs.pm/poolex/workers-and-callers-implementations.html)\n  - [Callers](https://hexdocs.pm/poolex/workers-and-callers-implementations.html#callers)\n  - [Workers](https://hexdocs.pm/poolex/workers-and-callers-implementations.html#workers)\n  - [Writing custom implementations](https://hexdocs.pm/poolex/workers-and-callers-implementations.html#writing-custom-implementations)","ref":"readme.html#guides"},{"type":"extras","title":"Contributions - Poolex","doc":"If you feel something can be improved or have any questions about specific behaviors or pieces of implementation, please feel free to file an issue. Proposed changes should be taken to issues before any PRs to save time on code that might not be merged upstream.\n\nIf you are ready to change the project, please read the [Contributing guide](docs/CONTRIBUTING.md) first.","ref":"readme.html#contributions"},{"type":"extras","title":"Contributing to Poolex","doc":"# Contributing to Poolex\n\nPoolex is written in [Elixir](https://elixir-lang.org/).\n\nFor branching management, this project uses [git-flow](https://github.com/petervanderdoes/gitflow-avh). The `main` branch is reserved for releases: the development process occurs on the `develop` and `feature` branches. Please never commit to `main`.\n\nYou can use [asdf](https://asdf-vm.com/) to set up the required Elixir and OTP. Current versions are listed in the `.tool-versions` file.","ref":"contributing.html"},{"type":"extras","title":"Setup - Contributing to Poolex","doc":"","ref":"contributing.html#setup"},{"type":"extras","title":"Local repository - Contributing to Poolex","doc":"1. Fork the repository.\n2. Clone your fork to a local repository:\n\n    ```shell\n    git clone https://github.com/your-login/poolex.git\n    cd poolex\n    ```\n\n3. Checkout `develop`:\n\n    ```shell\n    git checkout develop\n    ```","ref":"contributing.html#local-repository"},{"type":"extras","title":"Development environment (using asdf) - Contributing to Poolex","doc":"1. Install asdf by [Getting Started guideline](https://asdf-vm.com/guide/getting-started.html)\n2. Add plugins for elixir and OTP\n\n    ```shell\n    asdf plugin-add elixir https://github.com/asdf-vm/asdf-elixir.git\n    asdf plugin add erlang https://github.com/asdf-vm/asdf-erlang.git\n    ```\n\n3. Install tools:\n\n    ```shell\n    cd poolex\n    asdf install\n    ```","ref":"contributing.html#development-environment-using-asdf"},{"type":"extras","title":"Development environment (without asdf) - Contributing to Poolex","doc":"Please see [installation instructions](https://elixir-lang.org/install.html).","ref":"contributing.html#development-environment-without-asdf"},{"type":"extras","title":"Git-flow - Contributing to Poolex","doc":"If you want to use the `git-flow` CLI, please check [installation instructions](https://github.com/petervanderdoes/gitflow-avh/wiki/Installation).","ref":"contributing.html#git-flow"},{"type":"extras","title":"Building the project - Contributing to Poolex","doc":"1. Fetch the project dependencies:\n\n    ```shell\n    cd poolex\n    mix deps.get\n    ```\n\n2. Run the static analyzers:\n\n    ```shell\n    mix check\n    ```","ref":"contributing.html#building-the-project"},{"type":"extras","title":"Workflow - Contributing to Poolex","doc":"To make a change, please use this workflow:\n\n1. Checkout `develop` and apply the last upstream changes (use rebase, not merge!):\n\n    ```shell\n    git checkout develop\n    git fetch --all --prune\n    git rebase upstream/develop\n    ```\n\n2. For a tiny patch, create a new branch with an explicit name:\n\n    ```shell\n    git checkout -b  \n    ```\n\n    Alternatively, if you are working on a feature that would need more work, you can create a feature branch with `git-flow`:\n\n    ```shell\n    git flow feature start  \n    ```\n\n    *Note: always open an issue and ask before starting a big feature, to avoid it not being merged and your time lost.*\n\n3. When your feature is ready, feel free to use [interactive rebase](https://help.github.com/articles/about-git-rebase/) so your history looks clean and easy to follow. Then, apply the last upstream changes on `develop` to prepare integration:\n\n    ```shell\n    git checkout develop\n    git fetch --all --prune\n    git rebase upstream/develop\n    ```\n\n4. If there were commits on `develop` since the beginning of your feature branch, integrate them by **rebasing** if your branch has few commits, or merging if you had a long-lived branch:\n\n    ```shell\n    git checkout  \n    git rebase develop\n    ```\n\n    *Note: the only case you should merge is when you are working on a big feature. If it is the case, we should have discussed this before as stated above.*\n\n5. Run the tests and static analyzers to ensure there is no regression and all works as expected:\n\n    ```shell\n    mix check\n    ```\n\n6. If it’s all good, open a pull request to merge your branch into the `develop` branch on the main repository.","ref":"contributing.html#workflow"},{"type":"extras","title":"Coding style - Contributing to Poolex","doc":"Please format your code with `mix format` or your editor and follow\n[this style guide](https://github.com/christopheradams/elixir_style_guide).\n\nAll contributed code must be documented and functions must have typespecs. In general, take your inspiration from the existing code.","ref":"contributing.html#coding-style"},{"type":"extras","title":"Getting Started","doc":"# Getting Started","ref":"getting-started.html"},{"type":"extras","title":"Starting pool of workers - Getting Started","doc":"To start a pool you can use either `start/1` or `start_link/1`.\n\n```elixir\nPoolex.start_link(worker_module: SomeWorker, workers_count: 10)\n```\n\nIn general, you should place it into your Supervision tree for fault tolerance.\n\n```elixir\nchildren = [\n  {Poolex,\n  worker_module: SomeWorker,\n  workers_count: 10,\n  max_overflow: 10}\n]\n\nSupervisor.start_link(children, strategy: :one_for_one)\n```","ref":"getting-started.html#starting-pool-of-workers"},{"type":"extras","title":"Poolex configuration options - Getting Started","doc":"| Option                 | Description                                          | Example               | Default value                     |\n|------------------------|------------------------------------------------------|-----------------------|-------------------------|\n| `pool_id`              | Identifier by which you will access the pool         | `:my_pool`            | `worker_module` value                    |\n| `worker_module`        | Name of module that implements our worker            | `MyApp.Worker`        | **option is required**               |\n| `workers_count`        | How many workers should be running in the pool       | `5`                   | **option is required**               |\n| `max_overflow`         | How many workers can be created over the limit       | `2`                   | `0`                      |\n| `worker_args`          | List of arguments passed to the start function       | `[:gg, \"wp\"]`         | `[]`                        |\n| `worker_start_fun`     | Name of the function that starts the worker          | `:run`                | `:start_link`            |\n| `busy_workers_impl`    | Module that describes how to work with busy workers  | `SomeBusyWorkersImpl` | `Poolex.Workers.Impl.List`                    |\n| `idle_workers_impl`    | Module that describes how to work with idle workers  | `SomeIdleWorkersImpl` | `Poolex.Workers.Impl.List`                    |\n| `waiting_callers_impl` | Module that describes how to work with callers queue | `WaitingCallersImpl`  | `Poolex.Callers.Impl.ErlangQueue`             |\n| `pool_size_metrics`    | Whether to dispatch pool size metrics                | `true`                | `false`                  |","ref":"getting-started.html#poolex-configuration-options"},{"type":"extras","title":"Working with the pool - Getting Started","doc":"After the pool is initialized, you can get a free worker and perform any operations on it. This is done through the main interface `run/3`.\n\nThe first argument is the name of the pool mentioned above.\n\nThe second argument is the function that takes the pid of the worker as the only parameter and performs the necessary actions.\n\nThe third argument contains run options. Currently, there is only one `checkout_timeout` option that tells Poolex how long we can wait for a worker on the call site.\n\n```elixir\niex> Poolex.start_link(pool_id: :agent_pool, worker_module: Agent, worker_args: [fn -> 5 end], workers_count: 1)\niex> Poolex.run(:agent_pool, fn pid -> Agent.get(pid, &(&1)) end, checkout_timeout: 5_000)\n{:ok, 5}\n```\n\nIf you would like to see examples of using Poolex, then check out [Example of Use](https://hexdocs.pm/poolex/example-of-use.html).","ref":"getting-started.html#working-with-the-pool"},{"type":"extras","title":"Migration from poolboy","doc":"# Migration from `poolboy`\n\nIf you are using `:poolboy` and want to use `Poolex` instead, then you need to follow three simple steps.","ref":"migration-from-poolboy.html"},{"type":"extras","title":"Installation steps - Migration from poolboy","doc":"","ref":"migration-from-poolboy.html#installation-steps"},{"type":"extras","title":"I. Install the Poolex dependency - Migration from poolboy","doc":"#### mix.exs\n\n```diff\ndefp deps do\n  [\n-    {:poolboy, \"~> 1.5.0\"}\n+    {:poolex, \"~> 1.0\"}\n  ]\nend\n```\n\nInstall it.\n\n```bash\nmix deps.get\n```\n\nWell, you can also clean up installed dependencies locally and remove them from the `lock` file.\n\n```bash\nmix deps.clean --unlock --unused\n```","ref":"migration-from-poolboy.html#i-install-the-poolex-dependency"},{"type":"extras","title":"II. Update child specs - Migration from poolboy","doc":"#### Your Application or Supervisor file\n\n```diff\ndef init(_args) do\n  children = [\n-    :poolboy.child_spec(:some_pool,\n-      name: {:local, :some_pool},\n-      worker_module: MyApp.SomeWorker,\n-      size: 100,\n-      max_overflow: 50\n-    )\n+    {Poolex,\n+      pool_id: {:local, :some_pool},\n+      worker_module: MyApp.SomeWorker,\n+      workers_count: 100,\n+      max_overflow: 50}\n  ]\n\n  Supervisor.init(children, strategy: :one_for_one)\nend\n```","ref":"migration-from-poolboy.html#ii-update-child-specs"},{"type":"extras","title":"III. Update call site - Migration from poolboy","doc":"Use `run/3`.\nBe careful, unlike `:poolboy.transaction`, `Poolex.run` returns `{:ok, result}`.\n\n```diff\n-  result = :poolboy.transaction(\n-    :some_pool,\n-    fn pid -> some_function(pid) end,\n-    :timer.seconds(10)\n-  )\n+  {:ok, result} = Poolex.run(\n+    :some_pool,\n+    fn pid -> some_function(pid) end,\n+    checkout_timeout: :timer.seconds(10)\n+  )\n```","ref":"migration-from-poolboy.html#iii-update-call-site"},{"type":"extras","title":"Example of use","doc":"# Example of use\n\nThis example is based on the [Elixir School's poolboy guide](https://elixirschool.com/en/lessons/misc/poolboy).  \nYou can find the source of the below example here: [poolex_example](https://github.com/general-CbIC/poolex/tree/develop/examples/poolex_example).","ref":"example-of-use.html"},{"type":"extras","title":"Defining the worker - Example of use","doc":"We describe an actor that can easily become a bottleneck in our application since it has a rather long execution time on a blocking call.\n\n```elixir\ndefmodule PoolexExample.Worker do\n  use GenServer\n\n  def start_link do\n    GenServer.start_link(__MODULE__, nil)\n  end\n\n  def init(_args) do\n    {:ok, nil}\n  end\n\n  def handle_call({:square_root, x}, _from, state) do\n    IO.puts(\"process #{inspect(self())} calculating square root of #{x}\")\n    Process.sleep(1_000)\n    {:reply, :math.sqrt(x), state}\n  end\nend\n```","ref":"example-of-use.html#defining-the-worker"},{"type":"extras","title":"Configuring Poolex - Example of use","doc":"```elixir\ndefmodule PoolexExample.Application do\n  @moduledoc false\n\n  use Application\n\n  def start(_type, _args) do\n    children = [\n      {Poolex,\n       worker_module: PoolexExample.Worker,\n       workers_count: 5,\n       max_overflow: 2}\n    ]\n\n    Supervisor.start_link(children, strategy: :one_for_one)\n  end\nend\n```","ref":"example-of-use.html#configuring-poolex"},{"type":"extras","title":"Using Poolex - Example of use","doc":"`Poolex.run/3` is the function that you can use to interface with the worker pool.\n\n```elixir\ndefmodule PoolexExample.Test do\n  @timeout 60_000\n\n  def start do\n    1..20\n    |> Enum.map(fn i -> async_call_square_root(i) end)\n    |> Enum.each(fn task -> await_and_inspect(task) end)\n  end\n\n  defp async_call_square_root(i) do\n    Task.async(fn ->\n      Poolex.run(\n        PoolexExample.Worker,\n        fn pid ->\n          # Let's wrap the genserver call in a try-catch block. This allows us to trap any exceptions\n          # that might be thrown and return the worker to Poolex in a clean manner. It also allows\n          # the programmer to retrieve the error and potentially fix it.\n          try do\n            GenServer.call(pid, {:square_root, i})\n          catch\n            e, r ->\n              IO.inspect(\"Poolex transaction caught error: #{inspect(e)}, #{inspect(r)}\")\n              :ok\n          end\n        end,\n        checkout_timeout: @timeout\n      )\n    end)\n  end\n\n  defp await_and_inspect(task), do: task |> Task.await(@timeout) |> IO.inspect()\nend\n```\n\nRun the test function `PoolexExample.Test.start()` and see the result.\n\nNote that 7 workers started at once. 5 of them were launched at initialization and 2 workers were started over the limit based on the `max_overflow` setting.\n\nAll supported configuration options are presented in [Getting Started guide](https://hexdocs.pm/poolex/getting-started.html#poolex-configuration-options).\n\n```text\nprocess #PID<0.351.0> calculating square root of 3\nprocess #PID<0.350.0> calculating square root of 2\nprocess #PID<0.349.0> calculating square root of 1\nprocess #PID<0.352.0> calculating square root of 4\nprocess #PID<0.353.0> calculating square root of 5\nprocess #PID<0.349.0> calculating square root of 9\nprocess #PID<0.353.0> calculating square root of 7\n{:ok, 1.0}\nprocess #PID<0.351.0> calculating square root of 8\nprocess #PID<0.352.0> calculating square root of 6\nprocess #PID<0.350.0> calculating square root of 10\n{:ok, 1.4142135623730951}\n{:ok, 1.7320508075688772}\n...\n```","ref":"example-of-use.html#using-poolex"},{"type":"extras","title":"Working with metrics","doc":"# Working with metrics\n\nYou can quickly analyze and optimize your pool's production settings with the metrics presented by the library.","ref":"pool-metrics.html"},{"type":"extras","title":"Pool size metrics - Working with metrics","doc":"The Poolex library presents **an idle/busy worker count metric**. These metrics help estimate a pool load and the number of workers used.\n\nAlso, there is **an overflow metric**. It shows how long pools are forced to use additional workers.\n\nTo enable pool size metrics, you need to set the pool_size_metrics parameter to true on the pool initialization:\n\n```elixir\nchildren = [\n  {Poolex, \n    worker_module: SomeWorker,\n    workers_count: 5,\n    pool_size_metrics: true}\n]\n```\n\nYou can handle them by using `:telemetry.attach/4`:\n\n```elixir\n:telemetry.attach(\n  \"my-lovely-pool-size-metrics\",\n  [:poolex, :metrics, :pool_size],\n  &MyApp.handle_event/4,\n  nil\n)\n```\n\nFor example, your application can write metrics to the console: [PoolexExample.MetricsHandler](https://github.com/general-CbIC/poolex/blob/develop/examples/poolex_example/lib/poolex_example/metrics_handler.ex).\n\nMore about using `telemetry` [here](https://hexdocs.pm/telemetry/readme.html).","ref":"pool-metrics.html#pool-size-metrics"},{"type":"extras","title":"Integration with PromEx - Working with metrics","doc":"There is a plugin that works with the [PromEx](https://github.com/akoutmos/prom_ex) library: [Poolex.PromEx](https://hex.pm/packages/poolex_prom_ex).\n\nAbout installation of this plugin you can read [here](https://hexdocs.pm/poolex_prom_ex/readme.html#installation).","ref":"pool-metrics.html#integration-with-promex"},{"type":"extras","title":"Workers and callers implementations","doc":"# Workers and callers implementations\n\n`Poolex` operates with two concepts: `callers` and `workers`. In both cases, we are talking about processes.","ref":"workers-and-callers-implementations.html"},{"type":"extras","title":"Callers - Workers and callers implementations","doc":"`Callers` are processes that have requested to get a worker (used `run/3`). Each pool keeps the information about `callers` to distribute workers to them when they are free.\n\n>","ref":"workers-and-callers-implementations.html#callers"},{"type":"extras","title":"Caller's typespec {: .warning} - Workers and callers implementations","doc":">\n> Caller's typespec is `GenServer.from()` not a `pid()`.\n\nThe implementation of the caller storage structure should be conceptually similar to a queue since by default we want to give workers in the order they are requested. But this logic can be easily changed by writing your implementation.\n\nBehaviour of callers collection described [here](https://hexdocs.pm/poolex/Poolex.Callers.Behaviour.html).","ref":"workers-and-callers-implementations.html#caller-s-typespec-warning"},{"type":"extras","title":"Behaviour callbacks - Workers and callers implementations","doc":"| Callback          | Description                                                                                                          |\n|-------------------|----------------------------------------------------------------------------------------------------------------------|\n| `init/0`          | Returns `state` (any data structure) which will be passed as the first argument to all other functions.              |\n| `add/2`           | Adds caller to `state` and returns a new state.                                                                      |\n| `empty?/1`        | Returns `true` if the `state` is empty, `false` otherwise.                                                           |\n| `pop/1`           | Removes one of the callers from `state` and returns it as `{caller, state}`. Returns `:empty` if the state is empty. |\n| `remove_by_pid/2` | Removes given caller by caller's pid from `state` and returns a new state.                                           |\n| `to_list/1`       | Returns list of callers.                                                                                             |","ref":"workers-and-callers-implementations.html#behaviour-callbacks"},{"type":"extras","title":"Callers implementations out of the box - Workers and callers implementations","doc":"| Module                            | Description                                    | Source                                                                                           | Default? |\n|-----------------------------------|------------------------------------------------|--------------------------------------------------------------------------------------------------|----------|\n| `Poolex.Callers.Impl.ErlangQueue` | `FIFO` implementation based on `:erlang.queue` | [link](https://github.com/general-CbIC/poolex/blob/main/lib/poolex/callers/impl/erlang_queue.ex) | ✅       |","ref":"workers-and-callers-implementations.html#callers-implementations-out-of-the-box"},{"type":"extras","title":"Workers - Workers and callers implementations","doc":"**Workers** are processes launched in a pool. `Poolex` works with two collections of workers:\n\n1. `IdleWorkers` -- Free processes that can be given to callers upon request.\n2. `BusyWorkers` -- Processes that are currently processing the caller's request.\n\nFor both cases, the default implementation is based on lists. But it is possible to set different implementations for them.\n\nBehaviour of workers collection described [here](https://hexdocs.pm/poolex/Poolex.Workers.Behaviour.html).","ref":"workers-and-callers-implementations.html#workers"},{"type":"extras","title":"Behaviour callbacks - Workers and callers implementations","doc":"| Callback    | Description                                                                                                      |\n|-------------|------------------------------------------------------------------------------------------------------------------|\n| `init/0`    | Returns `state` (any data structure) which will be passed as the first argument to all other functions.          |\n| `init/1`    | Same as `init/0` but returns `state` initialized with a passed list of workers.                                  |\n| `add/2`     | Adds worker's pid to `state` and returns a new state.                                                            |\n| `member?/2` | Returns `true` if given worker contained in the `state`, `false` otherwise.                                      |\n| `remove/2`  | Removes given worker from `state` and returns new state.                                                         |\n| `count/1`   | Returns the number of workers in the state.                                                                      |\n| `to_list/1` | Returns list of workers pids.                                                                                    |\n| `empty?/1`  | Returns `true` if the `state` is empty, `false` otherwise.                                                       |\n| `pop/1`     | Removes one of workers from `state` and returns it as `{caller, state}`. Returns `:empty` if the state is empty. |","ref":"workers-and-callers-implementations.html#behaviour-callbacks"},{"type":"extras","title":"Workers implementations out of the box - Workers and callers implementations","doc":"| Module                            | Description                                    | Source                                                                                           | Default? |\n|-----------------------------------|------------------------------------------------|--------------------------------------------------------------------------------------------------|----------|\n| `Poolex.Workers.Impl.List`        | `LIFO` implementation based on Elixir's `List` | [link](https://github.com/general-CbIC/poolex/blob/main/lib/poolex/workers/impl/list.ex)         | ✅       |\n| `Poolex.Workers.Impl.ErlangQueue` | `FIFO` implementation based on `:erlang.queue` | [link](https://github.com/general-CbIC/poolex/blob/main/lib/poolex/workers/impl/erlang_queue.ex) | ❌       |","ref":"workers-and-callers-implementations.html#workers-implementations-out-of-the-box"},{"type":"extras","title":"Writing custom implementations - Workers and callers implementations","doc":"It's quite simple when using the [Behaviours](https://elixir-lang.org/getting-started/typespecs-and-behaviours.html#behaviours) mechanism in Elixir.\n\nFor example, you want to define a new implementation for callers. To do this, you need to create a module that inherits the [Poolex.Callers.Behaviour](../../lib/poolex/callers/behaviour.ex) and implement all its functions.\n\n```elixir\ndefmodule MyApp.MyAmazingCallersImpl do\n  @behaviour Poolex.Callers.Behaviour\n\n  def init, do: {}\n  def add(state, caller), do: #...\nend\n```\n\nIf you have any ideas about what implementations can be added to the library or how to improve existing ones, then please [create an issue](https://github.com/general-CbIC/poolex/issues/new)!","ref":"workers-and-callers-implementations.html#writing-custom-implementations"},{"type":"extras","title":"Configuring custom implementations - Workers and callers implementations","doc":"After that, you need to provide your module names to Poolex initialization:\n\n```elixir\nPoolex.child_spec(\n  pool_id: :some_pool,\n  worker_module: SomeWorker,\n  workers_count: 10,\n  waiting_callers_impl: MyApp.MyAmazingCallersImpl\n)\n```\n\nThat's it! Your implementation will be used in launched pool.\n\nThe configuration for workers might look like this:\n\n```elixir\nPoolex.child_spec(\n  pool_id: :some_pool,\n  worker_module: SomeWorker,\n  workers_count: 10,\n  busy_workers_impl: MyApp.PerfectBusyWorkersImpl,\n  idle_workers_impl: MyApp.FancyIdleWorkersImpl\n)\n```","ref":"workers-and-callers-implementations.html#configuring-custom-implementations"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.36.1"}}